from pathlib import Path
from typing import Dict, List

from dataclassy import dataclass

from ape.api.compiler import CompilerAPI
from ape.plugins import PluginManager
from ape.types import ContractType
from ape.utils import cached_property, notify

from .config import ConfigManager


@dataclass
class CompilerManager:
    config: ConfigManager
    plugin_manager: PluginManager

    @cached_property
    def registered_compilers(self) -> Dict[str, CompilerAPI]:
        registered_compilers = {}

        for plugin_name, (extensions, compiler_class) in self.plugin_manager.register_compiler:
            # TODO: Add config via ``self.config.get_config(plugin_name)``
            compiler = compiler_class()

            for extension in extensions:

                if extension in registered_compilers:
                    raise Exception("Extension already registered!")

                registered_compilers[extension] = compiler

        return registered_compilers

    def compile(self, contract_filepaths: List[Path]) -> Dict[str, ContractType]:
        extensions = set(path.suffix for path in contract_filepaths)
        if extensions > set(self.registered_compilers):
            raise Exception("No compiler found for extension")

        contract_types = {}
        for extension in extensions:
            paths_to_compile = [path for path in contract_filepaths if path.suffix == extension]
            for path in paths_to_compile:
                try:
                    notify("INFO", f"Compiling '{path.relative_to(self.config.PROJECT_FOLDER)}'")
                except ValueError:
                    notify("INFO", f"Compiling '{path}'")
            for contract_type in self.registered_compilers[extension].compile(paths_to_compile):

                if contract_type.contractName in contract_types:
                    raise Exception("ContractType collision across compiler plugins")

                contract_types[contract_type.contractName] = contract_type

        return contract_types



import pytest  # type: ignore
from ape_ethereum.converters import ETHER_UNITS
from eth_typing import ChecksumAddress
from hypothesis import given
from hypothesis import strategies as st

from ape import convert


@pytest.mark.fuzzing
@given(
    value=st.decimals(
        min_value=-(2 ** 255),
        max_value=2 ** 256 - 1,
        allow_infinity=False,
        allow_nan=False,
    ),
    unit=st.sampled_from(list(ETHER_UNITS.keys())),
)
def test_ether_conversions(value, unit):
    assert convert(value=f"{value} {unit}", type=int) == int(value * ETHER_UNITS[unit])


def test_bad_type():
    with pytest.raises(Exception):
        convert(value="something", type=float)


def test_no_registered_converter():
    with pytest.raises(Exception):
        convert(value="something", type=ChecksumAddress)



from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

from eth_utils import to_bytes

from ape.types import ABI, AddressType, ContractType
from ape.utils import notify

from .address import Address, AddressAPI
from .base import dataclass
from .providers import ProviderAPI, ReceiptAPI, TransactionAPI

if TYPE_CHECKING:
    from ape.managers.converters import ConversionManager
    from ape.managers.networks import NetworkManager


@dataclass
class ContractConstructor:
    deployment_bytecode: bytes
    abi: Optional[ABI]
    provider: ProviderAPI

    def __post_init__(self):
        if len(self.deployment_bytecode) == 0:
            raise Exception("No bytecode to deploy")

    def __repr__(self) -> str:
        return self.abi.signature if self.abi else "constructor()"

    def encode(self, *args, **kwargs) -> TransactionAPI:
        return self.provider.network.ecosystem.encode_deployment(
            self.deployment_bytecode, self.abi, *args, **kwargs
        )

    def __call__(self, *args, **kwargs) -> ReceiptAPI:
        if "sender" in kwargs:
            sender = kwargs["sender"]
            kwargs["sender"] = sender.address

            txn = self.encode(*args, **kwargs)
            return sender.call(txn)

        else:
            txn = self.encode(*args, **kwargs)
            return self.provider.send_transaction(txn)


@dataclass
class ContractCall:
    abi: ABI
    address: AddressType
    provider: ProviderAPI

    def __repr__(self) -> str:
        return self.abi.signature

    def encode(self, *args, **kwargs) -> TransactionAPI:
        return self.provider.network.ecosystem.encode_transaction(
            self.address, self.abi, *args, **kwargs
        )

    def __call__(self, *args, **kwargs) -> Any:
        txn = self.encode(*args, **kwargs)

        if "sender" in kwargs and not isinstance(kwargs["sender"], str):
            txn.sender = kwargs["sender"].address

        raw_output = self.provider.send_call(txn)
        tuple_output = self.provider.network.ecosystem.decode_calldata(  # type: ignore
            self.abi,
            raw_output,
        )

        # NOTE: Returns a tuple, so make sure to handle all the cases
        if len(tuple_output) < 2:
            return tuple_output[0] if len(tuple_output) == 1 else None

        else:
            # TODO: Handle struct output
            return tuple_output


@dataclass
class ContractCallHandler:
    provider: ProviderAPI
    address: AddressType
    abis: List[ABI]

    def __repr__(self) -> str:
        abis = sorted(self.abis, key=lambda abi: len(abi.inputs))
        return abis[-1].signature

    def __call__(self, *args, **kwargs) -> Any:
        selected_abi = None
        for abi in self.abis:
            if len(args) == len(abi.inputs):
                selected_abi = abi

        if not selected_abi:
            raise Exception("Number of args does not match")

        return ContractCall(  # type: ignore
            abi=selected_abi,
            address=self.address,
            provider=self.provider,
        )(*args, **kwargs)


@dataclass
class ContractTransaction:
    abi: ABI
    address: AddressType
    provider: ProviderAPI

    def __repr__(self) -> str:
        return self.abi.signature

    def encode(self, *args, **kwargs) -> TransactionAPI:
        return self.provider.network.ecosystem.encode_transaction(
            self.address, self.abi, *args, **kwargs
        )

    def __call__(self, *args, **kwargs) -> ReceiptAPI:

        if "sender" in kwargs:
            sender = kwargs["sender"]
            kwargs["sender"] = sender.address

            txn = self.encode(*args, **kwargs)
            return sender.call(txn)

        else:
            raise Exception("Must specify a `sender`")


@dataclass
class ContractTransactionHandler:
    provider: ProviderAPI
    address: AddressType
    abis: List[ABI]

    def __repr__(self) -> str:
        abis = sorted(self.abis, key=lambda abi: len(abi.inputs))
        return abis[-1].signature

    def __call__(self, *args, **kwargs) -> ReceiptAPI:
        selected_abi = None
        for abi in self.abis:
            if len(args) == len(abi.inputs):
                selected_abi = abi

        if not selected_abi:
            raise Exception("Number of args does not match")

        return ContractTransaction(  # type: ignore
            abi=selected_abi,
            address=self.address,
            provider=self.provider,
        )(*args, **kwargs)


@dataclass
class ContractLog:
    name: str
    data: Dict[str, Any]


@dataclass
class ContractEvent:
    provider: ProviderAPI
    address: str
    abis: List[ABI]
    cached_logs: List[ContractLog] = []


class ContractInstance(AddressAPI):
    _address: AddressType
    _contract_type: ContractType

    def __repr__(self) -> str:
        return f"<{self._contract_type.contractName} {self.address}>"

    @property
    def address(self) -> AddressType:
        return self._address

    def __dir__(self) -> List[str]:
        # This displays methods to IPython on `c.[TAB]` tab completion
        return list(super(AddressAPI, self).__dir__()) + [
            abi.name for abi in self._contract_type.abi
        ]

    def __getattr__(self, attr_name: str) -> Any:
        handlers = {
            "events": ContractEvent,
            "calls": ContractCallHandler,
            "transactions": ContractTransactionHandler,
        }

        def get_handler(abi_type: str) -> Any:
            selected_abis = [
                abi for abi in getattr(self._contract_type, abi_type) if abi.name == attr_name
            ]

            if not selected_abis:
                return  # No ABIs found for this type

            kwargs = {
                "provider": self.provider,
                "address": self.address,
                "abis": selected_abis,
            }

            try:
                return handlers[abi_type](**kwargs)  # type: ignore

            except Exception as e:
                # NOTE: Just a hack, because `__getattr__` *must* raise `AttributeError`
                raise AttributeError from e

        # Reverse search for the proper handler for this ABI name, if one exists
        for abi_type in handlers:
            handler = get_handler(abi_type)
            if handler:
                return handler
            # else: No ABI found with `attr_name`

        # No ABIs w/ name `attr_name` found at all
        raise AttributeError(f"{self.__class__.__name__} has no attribute '{attr_name}'")


@dataclass
class ContractContainer:
    _provider: ProviderAPI
    _contract_type: ContractType

    def __repr__(self) -> str:
        return f"<{self._contract_type.contractName}>"

    def at(self, address: str) -> ContractInstance:
        return ContractInstance(  # type: ignore
            _address=address,
            _provider=self._provider,
            _contract_type=self._contract_type,
        )

    @property
    def _deployment_bytecode(self) -> bytes:
        if (
            self._contract_type.deploymentBytecode
            and self._contract_type.deploymentBytecode.bytecode
        ):
            return to_bytes(hexstr=self._contract_type.deploymentBytecode.bytecode)

        else:
            return b""

    @property
    def _runtime_bytecode(self) -> bytes:
        if self._contract_type.runtimeBytecode and self._contract_type.runtimeBytecode.bytecode:
            return to_bytes(hexstr=self._contract_type.runtimeBytecode.bytecode)

        else:
            return b""

    def __call__(self, *args, **kwargs) -> TransactionAPI:
        constructor = ContractConstructor(  # type: ignore
            abi=self._contract_type.constructor,
            provider=self._provider,
            deployment_bytecode=self._deployment_bytecode,
        )
        return constructor.encode(*args, **kwargs)


def _Contract(
    address: Union[str, AddressAPI, AddressType],
    networks: "NetworkManager",
    converters: "ConversionManager",
    contract_type: Optional[ContractType] = None,
) -> AddressAPI:
    """
    Function used to triage whether we have a contract type available for
    the given address/network combo, or explicitly provided. If none are found,
    returns a simple ``Address`` instance instead of throwing (provides a warning)
    """

    # Check contract cache (e.g. previously deployed/downloaded contracts)
    # TODO: Add ``contract_cache`` dict-like object to ``NetworkAPI``
    # network = provider.network
    # if not contract_type and address in network.contract_cache:
    #    contract_type = network.contract_cache[address]

    # Check explorer API/cache (e.g. publicly published contracts)
    # TODO: Add ``get_contract_type`` to ``ExplorerAPI``
    # TODO: Store in ``NetworkAPI.contract_cache`` to reduce API calls
    # explorer = provider.network.explorer
    # if not contract_type and explorer:
    #    contract_type = explorer.get_contract_type(address)

    # We have a contract type either:
    #   1) explicity provided,
    #   2) from network cache, or
    #   3) from explorer
    if contract_type:
        return ContractInstance(  # type: ignore
            _address=converters.convert(address, AddressType),
            _provider=networks.active_provider,
            _contract_type=contract_type,
        )

    else:
        # We don't have a contract type from any source, provide raw address instead
        notify("WARNING", f"No contract type found for {address}")
        return Address(  # type: ignore
            _address=converters.convert(address, AddressType),
            _provider=networks.active_provider,
        )



from itertools import chain
from pathlib import Path

import click

from ape.utils import notify

flatten = chain.from_iterable


@click.command(short_help="Compile select contract source files")
@click.argument(
    "filepaths",
    nargs=-1,
    type=click.Path(exists=True, path_type=Path),
)
@click.option(
    "-f",
    "--force",
    "use_cache",
    flag_value=False,
    default=True,
    is_flag=True,
    help="Force recompiling selected contracts",
)
@click.option(
    "-s",
    "--size",
    "display_size",
    default=False,
    is_flag=True,
    help="Show deployment bytecode size for all contracts",
)
def cli(filepaths, use_cache, display_size):
    """
    Compiles the manifest for this project and saves the results
    back to the manifest.

    Note that ape automatically recompiles any changed contracts each time
    a project is loaded. You do not have to manually trigger a recompile.
    """
    # NOTE: Lazy load so that testing works properly
    from ape import project

    # Expand source tree based on selection
    if not filepaths:
        if not (project.path / "contracts").exists():
            notify("WARNING", "No `contracts/` directory detected")
            return

        # If no paths are specified, use all local project sources
        contract_filepaths = project.sources

    else:
        # Expand any folder paths
        expanded_filepaths = flatten([d.rglob("*.*") if d.is_dir() else [d] for d in filepaths])
        # Filter by what's in our project's source tree
        # NOTE: Make the paths absolute like `project.sources`
        contract_filepaths = [
            c.resolve() for c in expanded_filepaths if c.resolve() in project.sources
        ]

    if not contract_filepaths:
        if filepaths:
            selected_paths = "', '".join(str(p.resolve()) for p in filepaths)

        else:
            selected_paths = str(project.path / "contracts")

        extensions = ", ".join(set(f.suffix for f in selected_paths))
        notify("WARNING", f"No compilers detected for the following extensions: {extensions}")
        return

    # TODO: only compile selected contracts
    contract_types = project.load_contracts(use_cache)

    # Display bytecode size for *all* contract types (not just ones we compiled)
    if display_size:
        codesize = []
        for contract in contract_types.values():
            if not contract.deploymentBytecode:
                continue  # Skip if not bytecode to display

            bytecode = contract.deploymentBytecode.bytecode

            if bytecode:
                codesize.append((contract.contractName, len(bytecode) // 2))

        if not codesize:
            notify("INFO", "No contracts with bytecode to display")
            return

        click.echo()
        click.echo("============ Deployment Bytecode Sizes ============")
        indent = max(len(i[0]) for i in codesize)
        for name, size in sorted(codesize, key=lambda k: k[1], reverse=True):
            pct = size / 24577
            # pct_color = color(next((i[1] for i in CODESIZE_COLORS if pct >= i[0]), ""))
            # TODO Get colors fixed for bytecode size output
            # click.echo(f"  {name:<{indent}}  -  {size:>6,}B  ({pct_color}{pct:.2%}{color})")
            click.echo(f"  {name:<{indent}}  -  {size:>6,}B  ({pct:.2%})")

        click.echo()



import json
from pathlib import Path
from typing import Dict, List

from dataclassy import dataclass

from ape.api.config import ConfigDict, ConfigItem
from ape.plugins import PluginManager
from ape.utils import load_config

CONFIG_FILE_NAME = "ape-config.yaml"


@dataclass
class ConfigManager:
    DATA_FOLDER: Path
    REQUEST_HEADER: Dict
    PROJECT_FOLDER: Path
    name: str = ""
    version: str = ""
    dependencies: List[str] = []
    plugin_manager: PluginManager
    _plugin_configs: Dict[str, ConfigItem] = dict()

    def __post_init__(self):
        config_file = self.PROJECT_FOLDER / CONFIG_FILE_NAME

        if config_file.exists():
            user_config = load_config(config_file)
        else:
            user_config = {}

        # Top level config items
        self.name = user_config.pop("name", "")
        self.version = user_config.pop("version", "")
        self.dependencies = user_config.pop("dependencies", [])

        for plugin_name, config_class in self.plugin_manager.config_class:
            # NOTE: `dict.pop()` is used for checking if all config was processed
            user_override = user_config.pop(plugin_name, {})

            if config_class != ConfigDict:
                # NOTE: Will raise if improperly provided keys
                config = config_class(**user_override)

                # NOTE: Should raise if settings violate some sort of plugin requirement
                config.validate_config()

            else:
                # NOTE: Just use it directly as a dict if `ConfigDict` is passed
                config = user_override

            self._plugin_configs[plugin_name] = config

        if len(user_config.keys()) > 0:
            raise Exception("Unprocessed config items")

    def get_config(self, plugin_name: str) -> ConfigItem:
        if plugin_name not in self._plugin_configs:
            # plugin has no registered config class, so return empty config
            return ConfigItem()

        return self._plugin_configs[plugin_name]

    def serialize(self) -> Dict:
        project_config = dict()

        for name, config in self._plugin_configs.items():
            # NOTE: `config` is either `ConfigItem` or `dict`
            project_config[name] = config.serialize() if isinstance(config, ConfigItem) else config

        return project_config

    def __str__(self) -> str:
        return json.dumps(self.serialize())



import click

from ape import networks


@click.group(short_help="Manage networks")
def cli():
    """
    Command-line helper for managing networks.
    """


@cli.command(name="list", short_help="List registered networks")
def _list():
    click.echo("ecosystems:")
    default_ecosystem = networks.default_ecosystem.name
    for ecosystem_name in networks:
        if ecosystem_name == default_ecosystem:
            click.echo(f"- name: {ecosystem_name}  # Default")
        else:
            click.echo(f"- name: {ecosystem_name}")

        ecosystem = networks[ecosystem_name]

        default_network = ecosystem.default_network
        for network_name in getattr(networks, ecosystem_name):

            if network_name == default_network:
                click.echo(f"  - name: {network_name}  # Default")
            else:
                click.echo(f"  - name: {network_name}")

            network = ecosystem[network_name]

            if network.explorer:
                click.echo(f"    explorer: {network.explorer.name}")

            click.echo("    providers:")

            default_provider = network.default_provider
            for provider_name in network.providers:
                if provider_name == default_provider:
                    click.echo(f"    - {provider_name}  # Default")
                else:
                    click.echo(f"    - {provider_name}")



from typing import Dict

try:
    from importlib import metadata  # type: ignore
except ImportError:
    import importlib_metadata as metadata  # type: ignore

import click
import yaml

from ape.plugins import clean_plugin_name
from ape.utils import notify


def display_config(ctx, param, value):
    # NOTE: This is necessary not to interrupt how version or help is intercepted
    if not value or ctx.resilient_parsing:
        return

    from ape import project

    click.echo("# Current configuration")
    click.echo(yaml.dump(project.config.serialize()))

    ctx.exit()  # NOTE: Must exit to bypass running ApeCLI


class ApeCLI(click.MultiCommand):
    _commands = None

    @property
    def commands(self) -> Dict:
        if not self._commands:
            entry_points = metadata.entry_points()

            if "ape_cli_subcommands" not in entry_points:
                raise Exception("Missing registered cli subcommands")

            self._commands = {
                clean_plugin_name(entry_point.name): entry_point.load
                for entry_point in entry_points["ape_cli_subcommands"]
            }

        return self._commands

    def list_commands(self, ctx):
        return list(sorted(self.commands))

    def get_command(self, ctx, name):
        if name in self.commands:
            try:
                return self.commands[name]()
            except Exception:
                notify("WARNING", f"Error loading cli endpoint for plugin 'ape_{name}'")

        # NOTE: don't return anything so Click displays proper error


@click.command(cls=ApeCLI, context_settings=dict(help_option_names=["-h", "--help"]))
@click.version_option(message="%(version)s", package_name="eth-ape")
@click.option(
    "--config",
    is_flag=True,
    is_eager=True,
    expose_value=False,
    callback=display_config,
    help="Show configuration options (using `ape-config.yaml`)",
)
def cli():
    pass



import json
from typing import List

import click
from eth_account import Account as EthAccount  # type: ignore
from eth_utils import to_bytes

from ape import accounts
from ape.utils import notify

# NOTE: Must used the instantiated version of `AccountsContainer` in `accounts`
container = accounts.containers["accounts"]


class Alias(click.Choice):
    """Wraps ``click.Choice`` to load account aliases for the active project at runtime."""

    name = "alias"

    def __init__(self):
        # NOTE: we purposely skip the constructor of `Choice`
        self.case_sensitive = False

    @property
    def choices(self) -> List[str]:  # type: ignore
        # NOTE: This is a hack to lazy-load the aliases so CLI invocation works properly
        return list(accounts.aliases)


@click.group(short_help="Manage local accounts")
def cli():
    """
    Command-line helper for managing local accounts. You can unlock local accounts from
    scripts or the console using the accounts.load() method.
    """


# Different name because `list` is a keyword
@cli.command(name="list", short_help="List available accounts")
def _list():
    if len(accounts) == 0:
        notify("WARNING", "No accounts found.")
        return

    elif len(accounts) > 1:
        click.echo(f"Found {len(accounts)} accounts:")

    else:
        click.echo("Found 1 account:")

    for account in accounts:
        alias_display = f" (alias: '{account.alias}')" if account.alias else ""
        click.echo(f"  {account.address}{alias_display}")


@cli.command(short_help="Create a new keyfile account with a random private key")
@click.argument("alias")
def generate(alias):
    if alias in accounts.aliases:
        notify("ERROR", f"Account with alias '{alias}' already exists")
        return

    path = container.data_folder.joinpath(f"{alias}.json")
    extra_entropy = click.prompt(
        "Add extra entropy for key generation...",
        hide_input=True,
    )
    account = EthAccount.create(extra_entropy)
    passphrase = click.prompt(
        "Create Passphrase",
        hide_input=True,
        confirmation_prompt=True,
    )
    path.write_text(json.dumps(EthAccount.encrypt(account.key, passphrase)))

    notify("SUCCESS", f"A new account '{account.address}' has been added with the id '{alias}'")


# Different name because `import` is a keyword
@cli.command(name="import", short_help="Add a new keyfile account by entering a private key")
@click.argument("alias")
def _import(alias):
    if alias in accounts.aliases:
        notify("ERROR", f"Account with alias '{alias}' already exists")
        return

    path = container.data_folder.joinpath(f"{alias}.json")
    key = click.prompt("Enter Private Key", hide_input=True)
    try:
        account = EthAccount.from_key(to_bytes(hexstr=key))
    except Exception as error:
        notify("ERROR", f"Key can't be imported {error}")
        return
    passphrase = click.prompt(
        "Create Passphrase",
        hide_input=True,
        confirmation_prompt=True,
    )
    path.write_text(json.dumps(EthAccount.encrypt(account.key, passphrase)))

    notify("SUCCESS", f"A new account '{account.address}' has been added with the id '{alias}'")


@cli.command(short_help="Change the password of an existing account")
@click.argument("alias", type=Alias())
def change_password(alias):
    account = accounts.load(alias)
    account.change_password()
    notify("SUCCESS", f"Password has been changed for account '{alias}'")


@cli.command(short_help="Delete an existing account")
@click.argument("alias", type=Alias())
def delete(alias):
    account = accounts.load(alias)
    account.delete()
    notify("SUCCESS", f"Account '{alias}' has been deleted")
